//
// Adaptive Prediction Trees Decoder version 1.0
//
// Copyright (c) John Robinson. 1994, 1995, 1997, 2002, 2003, 2004.
//
// Converts APT to PGM or PPM using apt_decode() function.
// Also demonstrates use of apt_finish() to clean up.
//
#include <cstdlib>
#include <cstring>
#include "apt.h"
using apt1_0::apt_decode;
using apt1_0::apt_finish;

int writepnm(const char *pgmname, const unsigned char ***pbuf, 
const int& rows, const int& cols, const int& maxval, const int& pnmtype,
const int bytesperpel)
// Write picture to "P5" pgm or "P6" ppm format
// (If a "P4" format pbm picture was encoded, apt converts it into a
// two-level pgm picture.)
	{
	FILE *out = fopen(pgmname, "wb");
	if (!out) {
		fprintf(stderr,"Cannot open %s\n", pgmname);
		exit (0);
		}
	if (pnmtype == 1) {	// PGM
	    fprintf(out,"P5\n%d %d\n%d\n", cols, rows, maxval);
	    for (int i = 0; i < rows; i++)
		fwrite(pbuf[0][i],bytesperpel,cols,out);
	    }
	else { // PPM
	    fprintf(out,"P6\n%d %d\n%d\n", cols, rows, maxval);
	    unsigned char *temp = new unsigned char[3*cols*bytesperpel];
	    for (int i = 0; i < rows; i++) {
		register unsigned char *ptemp = temp;
		register const unsigned char *pg = pbuf[0][i];
		register const unsigned char *pr = pbuf[1][i];
		register const unsigned char *pb = pbuf[2][i];
		if (bytesperpel == 2) {
		    for (int j = 0; j < cols; j++) {
			*ptemp++ = *pr++;
			*ptemp++ = *pr++;
			*ptemp++ = *pg++;
			*ptemp++ = *pg++;
			*ptemp++ = *pb++;
			*ptemp++ = *pb++;
			}
		    }
		else {
		    for (int j = 0; j < cols; j++) {
			*ptemp++ = *pr++;
			*ptemp++ = *pg++;
			*ptemp++ = *pb++;
			}
		    }
		fwrite(temp,bytesperpel,cols*3,out);
		}
	    delete [] temp;
	    }
	fclose(out);
	return 0;
	}

int main(int argc, const char *argv[])
	{
	if (argc != 3) {
		fprintf(stderr, "Usage: dapt code_file pgm_outpic\n");
		return -1;
		}
	unsigned char **components[3];
	char *comment;
	int rows, cols;
	int pnmtype;
	int maxval;
	int retval;
	int bytesperpel;
	FILE *cfp = fopen(argv[1],"rb");
	if (!cfp) 
		retval = -1;
	else
		retval = apt_decode(components, cfp, rows, cols, 
			pnmtype, maxval,&comment, bytesperpel);
	switch (retval) {
		case -1:
			fprintf(stderr, "Can't open %s\n", argv[1]);
			return -1;
		case -2:
			fprintf(stderr, "%s is not a valid apt1.0 file\n", argv[1]);
			fclose(cfp);
			return -2;	// Not a apt file
		case -3:
			fprintf(stderr, "%s is coded with a later version of apt. Cannot decode.\n", argv[1]);
			fclose(cfp);
			return -3;	// Not a apt file
		default:
			fclose(cfp);
			break;
		}
	writepnm(argv[2], (const unsigned char ***) components, 
		rows, cols, maxval,pnmtype,bytesperpel);
	apt_finish(components, rows, cols, pnmtype);
	if (*comment)
		printf("Comment: %s\n",comment);
	return retval;
	}
		

