//
// Adaptive Prediction Trees
// Image Coding
//
// Version 1.0
//
// Based on...
//
// Binary tree predictive coding versions 1-5
//
// John Robinson 1994, 1995, 1997, 2002, 2003, 2004
// john@userport.com
//

// Order statistics prefilter for lossy encoding

namespace apt1_0 {
void order_filter(unsigned char ***components, int nr, int nc, int numcomps,
						int quality) {
	// 3x3 modified median filter. Messiness is in output buffering
	unsigned char *outbuf1 = new unsigned char[nc];
	unsigned char *outbuf2 = new unsigned char[nc];
	unsigned char sortarray[9];
	unsigned char sortarray2[9];
	int ovrange = 90 - quality;
	if (ovrange > 80)
		ovrange = 80;
	for (int compnum = 0; compnum < numcomps; compnum++) {
		int range = ovrange;
		if (compnum)
		 	range = ovrange*2;
		unsigned char *prevout = outbuf1;
		unsigned char *currout = outbuf2;
		unsigned char **pic = components[compnum];
		// Initialize first "previous" line
		memcpy(prevout, pic[0], nc);
		for (int i = 1; i < nr-1; i++) {
			// Initialize first and last points of current line
			currout[0] = pic[i][0];
			currout[nc-1] = pic[i][nc-1];
			for (int j = 1; j < nc-1; j++) {
				// Clunky for now, just to get working
				int maxindex = 0;
				int minindex = 0;
				int minval, maxval;
				int curr;
				minval = maxval = sortarray[0] = pic[i-1][j-1];
				curr = sortarray[1] = pic[i-1][j];
				if (curr > maxval) {maxval=curr; maxindex=1;}
				else if (curr<minval) {minval=curr; minindex=1;}
				curr = sortarray[2] = pic[i-1][j+1];
				if (curr > maxval) {maxval=curr; maxindex=2;}
				else if (curr<minval) {minval=curr; minindex=2;}
				curr = sortarray[3] = pic[i][j-1];
				if (curr > maxval) {maxval=curr; maxindex=3;}
				else if (curr<minval) {minval=curr; minindex=3;}
				curr = sortarray[4] = pic[i][j];
				if (curr > maxval) {maxval=curr; maxindex=4;}
				else if (curr<minval) {minval=curr; minindex=4;}
				curr = sortarray[5] = pic[i][j+1];
				if (curr > maxval) {maxval=curr; maxindex=5;}
				else if (curr<minval) {minval=curr; minindex=5;}
				curr = sortarray[6] = pic[i+1][j-1];
				if (curr > maxval) {maxval=curr; maxindex=6;}
				else if (curr<minval) {minval=curr; minindex=6;}
				curr = sortarray[7] = pic[i+1][j];
				if (curr > maxval) {maxval=curr; maxindex=7;}
				else if (curr<minval) {minval=curr; minindex=7;}
				curr = sortarray[8] = pic[i+1][j+1];
				if (curr > maxval) {maxval=curr; maxindex=8;}
				else if (curr<minval) {minval=curr; minindex=8;}
				if (maxval - minval > range) {
					currout[j] = pic[i][j];
					continue;
					}
				int ii, jj;
				for (ii = 0, jj = 0; ii < 9; ii++) {
					if ((ii == minindex)||(ii == maxindex))
						continue;
					sortarray2[jj++] = sortarray[ii];
					}
				minindex = maxindex = 0;
				minval = maxval = sortarray2[0];
				curr = sortarray2[1];
				if (curr > maxval) {maxval=curr; maxindex=1;}
				else if (curr<minval) {minval=curr; minindex=1;}
				curr = sortarray2[2];
				if (curr > maxval) {maxval=curr; maxindex=2;}
				else if (curr<minval) {minval=curr; minindex=2;}
				curr = sortarray2[3];
				if (curr > maxval) {maxval=curr; maxindex=3;}
				else if (curr<minval) {minval=curr; minindex=3;}
				curr = sortarray2[4];
				if (curr > maxval) {maxval=curr; maxindex=4;}
				else if (curr<minval) {minval=curr; minindex=4;}
				curr = sortarray2[5];
				if (curr > maxval) {maxval=curr; maxindex=5;}
				else if (curr<minval) {minval=curr; minindex=5;}
				curr = sortarray2[6];
				if (curr > maxval) {maxval=curr; maxindex=6;}
				else if (curr<minval) {minval=curr; minindex=6;}
				for (ii = 0, jj = 0; ii < 7; ii++) {
					if ((ii == minindex)||(ii == maxindex))
						continue;
					sortarray[jj++] = sortarray2[ii];
					}
				minindex = maxindex = 0;
				minval = maxval = sortarray[0];
				curr = sortarray[1];
				if (curr > maxval) {maxval=curr; maxindex=1;}
				else if (curr<minval) {minval=curr; minindex=1;}
				curr = sortarray[2];
				if (curr > maxval) {maxval=curr; maxindex=2;}
				else if (curr<minval) {minval=curr; minindex=2;}
				curr = sortarray[3];
				if (curr > maxval) {maxval=curr; maxindex=3;}
				else if (curr<minval) {minval=curr; minindex=3;}
				curr = sortarray[4];
				if (curr > maxval) {maxval=curr; maxindex=4;}
				else if (curr<minval) {minval=curr; minindex=4;}
				if (range < 40) {
					if (pic[i][j] > maxval)
						currout[j] = maxval;
					else if (pic[i][j] < minval)
						currout[j] = minval;
					else
						currout[j] = pic[i][j];
					continue;
					}
				for (ii = 0, jj = 0; ii < 5; ii++) {
					if ((ii == minindex)||(ii == maxindex))
						continue;
					sortarray2[jj++] = sortarray[ii];
					}
				minindex = maxindex = 0;
				minval = maxval = sortarray2[0];
				curr = sortarray2[1];
				if (curr > maxval) {maxval=curr; maxindex=1;}
				else if (curr<minval) {minval=curr; minindex=1;}
				curr = sortarray2[2];
				if (curr > maxval) {maxval=curr; maxindex=2;}
				else if (curr<minval) {minval=curr; minindex=2;}
				if (pic[i][j] > maxval)
					currout[j] = maxval;
				else if (pic[i][j] < minval)
					currout[j] = minval;
				else
					currout[j] = pic[i][j];
				}
			// At the end of the line we can write the previous
			// line, because won't be used for input again
			memcpy(pic[i-1],prevout,nc);
			if (prevout == outbuf1) {
				prevout = outbuf2;
				currout = outbuf1;
				}
			else {
				prevout = outbuf1;
				currout = outbuf2;
				}
			}
		// Write out the last line
		memcpy(pic[nr-2],prevout,nc);
		}
	delete [] outbuf1;
	delete [] outbuf2;
	}
};
