//
// Calculation of psnr between two PGM or PPM images
// John Robinson, 1995
//
#include <fstream>
#include <iostream>
#include <cmath>
#include <cstring>
#include <cstdlib>
using namespace std;
void pgmword(ifstream *pin, char *cbuf)
// Read next PGM word, ignoring comments and whitespace
	{
	char c;
	while(1) {
	    do 
		c = pin->get();
	    while(c == ' ' || c == '\t' || c == '\n' || c == '\r');
	    if (c == '#') {
		do
			c = pin->get();
		while (c != '\n');
		}
	    else break;
	    }
	*cbuf++ = c;
	while ((c = pin->peek())!= ' ' && c != '\t' && c != '\n' && c != '\r'
	  && c != '#')
	    *cbuf++ = pin->get();
	*cbuf = '\0';
	}
ifstream *readpgmheader(char *pgmname, int& rows, int& cols, int& comps,
		int& numbits)
// Read "P5" and "P6" pgm/ppm format header.
	{
	char cbuf[32];
	int maxval;
	unsigned char c;
	ifstream *pin = new ifstream(pgmname,ios::binary);
	if (pin->bad()) {
		cerr << "Cannot open " << pgmname << '\n';
		return 0;
		}
	cbuf[0] = pin->get();
	cbuf[1] = pin->get();
	cbuf[2] = 0;
	if (!strcmp(cbuf,"P5")) 
		comps = 1;
	else if (!strcmp(cbuf,"P6"))
		comps = 3;
	else {
		cerr << pgmname << " is not a valid raw format PGM/PPM file\n";
		return 0;
		}
	pgmword(pin, cbuf);
	cols = atoi(cbuf);
	pgmword(pin,cbuf);
	rows = atoi(cbuf);
	pgmword(pin,cbuf);
	maxval = atoi(cbuf);
	for (numbits = 0; numbits < 16; numbits++) {
		if (maxval < (1<<numbits))
			break;
		}
	// absorb final whitespace char before start of data
	c = pin->get();
	return pin;
	}

int main(int argc, char *argv[]) {
	if (argc != 3) {
		cerr << "Usage: psnr inpic outpic\n";
		return -1;
		}
	int rows, cols;
	int components;
	int numbits;
	int numbytes;
        unsigned char *inbuf;
        unsigned char *outbuf;
	ifstream *in, *out;
	in = readpgmheader(argv[1],rows,cols,components,numbits);
	if (!in)
		return 0;
	int temprows, tempcols, tempcomponents, tempnumbits;
	out=readpgmheader(argv[2],temprows,tempcols,tempcomponents,tempnumbits);
	if (!out)
		return 0;
	if ((tempcols != cols)||(temprows != rows)) {
		cerr << "Images are not the same size\n";
		return 0;
		}
	if (components != tempcomponents) {
		cerr << "One image PPM, one PGM: Cannot compare\n";
		return 0;
		}
	if (((numbits+7)/8) != ((tempnumbits+7)/8)) {
		cerr << "Sorry. This psnr program can't compare ";
		cerr << "pnm files where one picture has more than\n";
		cerr << "8 bits and the other has less.\n";
		return 0;
		}
	if (numbits < tempnumbits)
		numbits = tempnumbits;
	numbytes = (numbits+7)/8;
        inbuf = new unsigned char[cols*components*numbytes];
        outbuf = new unsigned char[cols*components*numbytes];
	double mse = 0.0;
	for (int i = 0; i < rows; i++) {
		unsigned char *p = inbuf;
		unsigned char *q = outbuf;
		int temp;
		in->read((char *)inbuf,cols*components*numbytes);
		out->read((char *)outbuf,cols*components*numbytes);
		for (int j = 0; j < cols*components; j++)
			{
			if (numbytes == 1) {
				temp = *p++ - *q++;
				mse += temp*temp;
				}
			else {
				temp = (*p++)<< 8;
				temp |= *p++;
				temp -= ((*q++)<<8);
				temp -= *q++;
				mse += temp*temp;
				}
			}
		}
	mse /= ((double)rows)*cols*components;
	if (mse == 0)
		cout << " Infinity\n";
	else {
		double psnr;
		double maxval;
		// We only treat 8, 12 and 16 bits maxima
		if (numbits <= 8)
			maxval = 255.0;
		else if (numbits <= 12)
			maxval = 4095.0;
		else
			maxval = 65535.0;
		psnr = maxval*maxval/mse;
		psnr = 10*log10(psnr);
		cout << ' ' << psnr << '\n';
		}
	in->close();
	out->close();
	delete in;
	delete out;
	return 1;
	}
